/*
 *  A_DICE Dice Rolling Simulator       Copyright (C) Kimmo Kulovesi 1997, 1998
 *                                                    arkhan@softhome.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  The R250 engine first written by W. L. Maier, as described by
 *  S. Kirkpatrick and E. Stoll, Journal of Computational Physics,
 *  40, p. 517 (1981). Modified by me.
 *
 *  Please see 'license.txt'.
 *
 ******************************************************************************

    Init_Dice(seed)           Initializes the random number generator with
                              an unsigned long seed, such as "time(NULL)".
                              "srand()" will not work with this random number
                              generator.

    RollDie(sides)            Simulates the rolling of a number of dice with
    RollDice(number, sides)   a specified number of sides each.
                                The number of sides must NEVER be 0, or it will
                              crash with a divide by zero. The values are given
                              as UNSIGNED ints, so do not attempt to give
                              negative values to the generator, or you might
                              experience odd errors.
                                About as fast or slightly faster than "rand()",
                              but these still give more random results.

    RDie(sides)               As above, but about 35% faster. The values given
    RDice(sides)              are slightly less "random", but probably does not
                              matter in any purposes which requires the speed.
                                Still way better and faster than "rand()".

    RollDie_B(sides)          Just as the ones above, without the "_B", but
    RollDice_B(number, sides) take the values as SIGNED ints, and can take
    RDie_B(sides              <= 0 values in "sides" without causing errors.
    RDice_B(sides)            In fact, if "sides" negative, it will return a
                              negative number randomized on the basis.
                                Number of dice can also be negative or 0, but
                              in either case it simply returns "0".
                                Because of the checks, they are, naturally,
                              slightly slower, but I definitely recommend these
                              if it is possible that "sides" might be 0...
***
            Also defined are the standard roleplaying dice =

        RollD2(),       1 -    2    (a coin flip, also; 1 = heads, 2 = tails)
        RollD3(),       1 -    3
        RollD4(),       1 -    4
        RollD6(),       1 -    6
        RollD8(),       1 -    8
        RollD10(),      1 -   10
        RollD12(),      1 -   12
        RollD20(),      1 -   20
        RollD100(),     1 -  100
        RollD1000(),    1 - 1000

            And the faster, but ever-so-slightly less random versions:

        RD2(),          1 -    2
        RD3(),          1 -    3
        RD4(),          1 -    4
        RD6(),          1 -    6
        RD8(),          1 -    8
        RD10(),         1 -   10
        RD12(),         1 -   12
        RD20(),         1 -   20
        RD100(),        1 -  100
        RD1000(),       1 - 1000
***
        Then, there is the raw R250 generator, which the rest of these are
based on:

        rgen()          Use this for a replacement for "rand()", since it is
                        about 35-45% faster, and gives better random numbers.

-----------------------------------------------------------------------------*/

#ifndef arkhan_a_dice_h
#define arkhan_a_dice_h

#define RollDie(n)   (1 + (((rgen() / 211) & 0777777) % (n)))
#define RollD2()     (1 + (((rgen() / 211) & 0777777) % 2))
#define RollD3()     (1 + (((rgen() / 211) & 0777777) % 3))
#define RollD4()     (1 + (((rgen() / 211) & 0777777) % 4))
#define RollD6()     (1 + (((rgen() / 211) & 0777777) % 6))
#define RollD8()     (1 + (((rgen() / 211) & 0777777) % 8))
#define RollD10()    (1 + (((rgen() / 211) & 0777777) % 10))
#define RollD12()    (1 + (((rgen() / 211) & 0777777) % 12))
#define RollD20()    (1 + (((rgen() / 211) & 0777777) % 20))
#define RollD100()   (1 + (((rgen() / 211) & 0777777) % 100))
#define RollD1000()  (1 + (((rgen() / 211) & 0777777) % 1000))

#define RDie(n)      (1 + rgen() % (n))
#define RD2()        (1 + rgen() % 2)
#define RD3()        (1 + rgen() % 3)
#define RD4()        (1 + rgen() % 4)
#define RD6()        (1 + rgen() % 6)
#define RD8()        (1 + rgen() % 8)
#define RD10()       (1 + rgen() % 10)
#define RD12()       (1 + rgen() % 12)
#define RD20()       (1 + rgen() % 20)
#define RD100()      (1 + rgen() % 100)
#define RD1000()     (1 + rgen() % 1000)

    /* And the prototypes : */

unsigned int  rgen();
void          Init_Dice (unsigned long seed);
unsigned long RollDice  (const unsigned int num_dice, const unsigned int num_sides);
long          RollDice_B(const int num_dice, const int num_sides);
int           RollDie_B (const int num_sides);
unsigned long RDice     (const unsigned int num_dice, const unsigned int num_sides);
long          RDice_B   (const int num_dice, const int num_sides);
int           RDie_B    (const int num_sides);

#endif
