/*
 *  A_WGEN Random Weather Generator     Copyright (C) Kimmo Kulovesi 1998
 *                                                    arkhan@softhome.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Please see 'license.txt'.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "a_dice.h"
#include "a_world.h"
#include "a_wgen.h"

/*****************************************************************************
  * Global variables and the structure definitions                           *
  ****************************************************************************/

typedef struct S_Weather
{
    unsigned char temperature,
                  weather,
                  wind_force;
    char          wind_direction;
}
S_Weather;

typedef struct WGEN
{
    unsigned char num[4];
    S_Weather     season[4][max_weather];
}
WGEN;


WGEN          *wgen         = NULL;
char          *weatherfile  = NULL;
signed char   wgen_position = -125;
int           wg_latitude;

/*****************************************************************************
  * wgen_generate: Get the next weather entry                                *
  ****************************************************************************/

Weather wgen_generate(unsigned char c_season, unsigned char is_at_night, unsigned char prev_precipitation)
{
    Weather w;
    unsigned char wgen_prev = 0;

    switch(prev_precipitation)
    {
        case p_thunder:
        case p_thunder_hail:
                            wgen_prev |= i_thunder;
                            wgen_prev |= i_heavy_rain;
                        break;
        case p_mist:
        case p_fog:
                            wgen_prev |= i_mist;
                        break;
        case p_rain:
                            wgen_prev |= i_rain;
                        break;
        case p_heavy_rain:
                            wgen_prev |= i_heavy_rain;
                        break;
        case p_snow:
                            wgen_prev |= i_snow;
                        break;
        case p_heavy_snow:
                            wgen_prev |= i_heavy_snow;
                        break;
    }

    if (wgen == NULL)
    {
        wgen_err(WGEN_ERR_NOT_READY, "Error: Tried to generate while WGEN isn't loaded.");

        w.wind_direction = dir_E;
        w.temperature = w.precipitation = w.cloud_cover = w.wind_force = 0;
        return w;
    }

        /* Generate initial position, if this is the first time run */

    if (wgen_position == -125)
        wgen_position = (RollDie(wgen[0].num[c_season]) - 1);

        /* Check that the position is correct for this season */

    else if (wgen_position >= wgen[0].num[c_season])
        wgen_position = 0;
    else if (wgen_position < 0)
        wgen_position = (wgen[0].num[c_season] - 1);

        /* Get the weather values */

    if (is_at_night)
    {
        if (wgen[0].season[c_season][wgen_position].temperature & tse_freezing)
            w.temperature = temperature_freezing;
        else if (wgen[0].season[c_season][wgen_position].temperature & tse_cold)
            w.temperature = temperature_cold;
        else if (wgen[0].season[c_season][wgen_position].temperature & tse_cool)
            w.temperature = temperature_cool;
        else if (wgen[0].season[c_season][wgen_position].temperature & tse_warm)
            w.temperature = temperature_warm;
        else if (wgen[0].season[c_season][wgen_position].temperature & tse_hot)
            w.temperature = temperature_hot;
        else
            w.temperature = wgen[0].season[c_season][wgen_position].temperature;
    }
    else
    {
        w.temperature = wgen[0].season[c_season][wgen_position].temperature;

            /* Remove any distracting secondary temperatures */

        w.temperature &= ~tse_freezing; w.temperature &= ~tse_cold;
        w.temperature &= ~tse_cool;     w.temperature &= ~tse_warm;
        w.temperature &= ~tse_hot;
    }

    w.wind_direction = wgen[0].season[c_season][wgen_position].wind_direction;
    w.wind_force = (RollDie(3) + wgen[0].season[c_season][wgen_position].wind_force) - 1;
    if (w.wind_force > 4)
        w.wind_force = 4;

    w.precipitation = p_none;
    if(wgen[0].season[c_season][wgen_position].weather & w_thunder)
    {
            /* Mark 'uncertain' thunder with the snow-flag */

        if (wgen[0].season[c_season][wgen_position].weather & w_snow)
        {
            if (!(wgen_prev & i_thunder) || RollDie(100) <= 15)
            {
                if (RollDie(100) <= 7)
                    w.precipitation = p_thunder_hail;
                else
                    w.precipitation = p_thunder;
            }
            else
            {
                if (wgen[0].season[c_season][wgen_position].weather & w_rain)
                {
                    if (wgen[0].season[c_season][wgen_position].weather & w_heavy)
                        w.precipitation = p_heavy_rain;
                    else
                        w.precipitation = p_rain;
                }
            }
        }
        else
        {
            if(wgen[0].season[c_season][wgen_position].weather & w_hail)
            {
                if (RollDie(100) <= 15)
                    w.precipitation = p_thunder_hail;
                else
                    w.precipitation = p_thunder;
            }
            else
            {
                if (RollDie(100) <= 7)
                    w.precipitation = p_thunder_hail;
                else
                    w.precipitation = p_thunder;
            }
        }
    }
    else if(wgen[0].season[c_season][wgen_position].weather & w_rain)
    {
        if (wgen[0].season[c_season][wgen_position].weather & w_heavy)
            w.precipitation = p_heavy_rain;
        else
            w.precipitation = p_rain;
    }
    else if(wgen[0].season[c_season][wgen_position].weather & w_snow)
    {
        switch(w.temperature)
        {
            case temperature_warm:
                                if (RollDie(100) <= 95)
                                    w.precipitation = p_sleet;
                            break;
            case temperature_cool:
                                if (RollDie(100) <= 40)
                                    w.precipitation = p_sleet;
                            break;
            case temperature_cold:
                                if (RollDie(100) <= 25)
                                    w.precipitation = p_sleet;
                            break;
            case temperature_hot:
                                w.precipitation = p_sleet;
            case temperature_freezing:
                            break;
        }

        if (w.precipitation == p_sleet)
        {
            if (wgen[0].season[c_season][wgen_position].weather & w_heavy)
            {
                w.precipitation = p_heavy_sleet;

                if (w.temperature < temperature_warm)
                {
                    if (wgen_prev & i_snow)
                    {
                        if (RollDie(100) <= 75)
                            w.precipitation = p_heavy_snow;
                    }
                    else if (wgen_prev & i_heavy_snow)
                    {
                        if (RollDie(100) <= 90)
                            w.precipitation = p_heavy_snow;
                    }
                }
            }
            else
            {
                if (w.temperature < temperature_warm)
                {
                    if (wgen_prev & i_snow)
                    {
                        if (RollDie(100) <= 80)
                            w.precipitation = p_snow;
                    }
                    else if (wgen_prev & i_heavy_snow)
                    {
                        if (RollDie(100) <= 95)
                            w.precipitation = p_snow;
                    }
                }
            }
        }

        if (w.precipitation != p_sleet && w.precipitation != p_heavy_sleet)
        {
            if (wgen[0].season[c_season][wgen_position].weather & w_heavy)
            {
                if (w.wind_force == 4)
                    w.precipitation = p_blizzard;
                else
                    w.precipitation = p_heavy_snow;

                if (w.temperature > temperature_freezing)
                {
                    if (wgen_prev & i_rain)
                    {
                        if (RollDie(100) <= 75)
                            w.precipitation = p_heavy_sleet;
                    }
                    else if (wgen_prev & i_heavy_rain)
                    {
                        if (RollDie(100) <= 90)
                            w.precipitation = p_heavy_sleet;
                    }
                }
            }
            else
            {
                w.precipitation = p_snow;
                if (w.temperature > temperature_freezing)
                {
                    if (wgen_prev & i_rain)
                    {
                        if (RollDie(100) <= 80)
                            w.precipitation = p_sleet;
                    }
                    else if (wgen_prev & i_heavy_rain)
                    {
                        if (RollDie(100) <= 95)
                            w.precipitation = p_sleet;
                    }
                }
            }
        }
    }
    else if(wgen[0].season[c_season][wgen_position].weather & w_mist)
    {
        switch(w.wind_force)
        {
            case 0:
                    w.precipitation = p_fog;
                break;
            case 1:
                    w.precipitation = p_mist;
                break;
            case 2:
                    if (wgen_prev & i_mist)
                    {
                        if (RollDie(100) <= 75)
                            w.precipitation = p_mist;
                    }
                    else
                    {
                        if (RollDie(100) <= 45)
                            w.precipitation = p_mist;
                    }
                break;
        }
    }
    else if(wgen[0].season[c_season][wgen_position].weather & w_hail)
    {
        if (RollDie(100) <= 3)
            w.precipitation = p_hail;
    }

    if (wgen[0].season[c_season][wgen_position].weather & w_cloudy)
        w.cloud_cover = RollDie(60) + 19;
    else if (wgen[0].season[c_season][wgen_position].weather & w_overcast)
        w.cloud_cover = RollDie(21) + 79;
    else
        w.cloud_cover = RollDie(20) - 1;

        /* Move the position marker */

    switch(RollDie(15))
    {
        case 1:
        case 2:
                wgen_position -= 1;
            break;
        case 12:
        case 13:
        case 14:
                wgen_position += 1;
            break;
        case 15:
                wgen_position += 2;
    }

    return w;
};

/*****************************************************************************
  * wgen_load: Read weather data into memory                                 *
  ****************************************************************************/

    /* Define the status flags */

#define st_spring       (season_spring)
#define st_summer       (season_summer)
#define st_autumn       (season_autumn)
#define st_winter       (season_winter)
#define st_info         10
#define st_location     11
#define st_latitude     12
#define st_none         13
#define st_quit         69 

int wgen_load(char *filename)
{
    FILE            *fle;
    static char     tempstr[256];
    char            wparm[9];
    int             i,
                    line = 0;
    unsigned char   status = st_none,
                    num_exceeded = 0,
                    j, k;

/*-----------------------------------------------------------------------------
  | Check if we are already using the file specified                          |
  ----------------------------------------------------------------------------*/

    if (weatherfile != NULL)
        if (!strcmp(weatherfile, filename))
            return 0;

/*-----------------------------------------------------------------------------
  | Open the file                                                             |
  ----------------------------------------------------------------------------*/

    fle = fopen(filename, "r");             /* Attempt to open as-is */

    if (fle == NULL)
    {
        strcpy(tempstr, filename);          /* Check if extension missing */
        strcat(tempstr, ".wth");
        fle = fopen(tempstr, "r");

        if (fle == NULL)
        {
            strcpy(tempstr, "Error: Requested weatherfile (");
            strcat(tempstr, filename);
            strcat(tempstr, ") cannot be opened.");
            wgen_err(WGEN_ERR_NO_FILE, tempstr);
            return -1;
        }
    }

    wgen_exit();        /* Reset WGEN */

/*-----------------------------------------------------------------------------
  | Now read through the file                                                 |
  ----------------------------------------------------------------------------*/

    wgen        = (WGEN *)malloc(sizeof(WGEN));
    weatherfile = (char *)malloc(strlen(filename) + 1);

    if (weatherfile == NULL || wgen == NULL)
    {
        wgen_err(WGEN_ERR_NO_MEMORY_LOAD, "Error: Out of memory loading weatherfile.");
        return -1;
    }

    strcpy(weatherfile, filename);

    i = 0;
    while (i < max_weather)
    {
        wgen[0].season[season_spring][i].weather = 0;
        wgen[0].season[season_spring][i].temperature = 0;
        wgen[0].season[season_spring][i].wind_force = 0;
        wgen[0].season[season_spring][i].wind_direction = dir_E;
        wgen[0].season[season_summer][i].weather = 0;
        wgen[0].season[season_summer][i].temperature = 0;
        wgen[0].season[season_summer][i].wind_force = 0;
        wgen[0].season[season_summer][i].wind_direction = dir_E;
        wgen[0].season[season_autumn][i].weather = 0;
        wgen[0].season[season_autumn][i].temperature = 0;
        wgen[0].season[season_autumn][i].wind_force = 0;
        wgen[0].season[season_autumn][i].wind_direction = dir_E;
        wgen[0].season[season_winter][i].weather = 0;
        wgen[0].season[season_winter][i].temperature = 0;
        wgen[0].season[season_winter][i].wind_force = 0;
        wgen[0].season[season_winter][i].wind_direction = dir_E;
        i++;
    }
    wgen[0].num[season_spring] = wgen[0].num[season_summer] = 0;
    wgen[0].num[season_autumn] = wgen[0].num[season_winter] = 0;
    wg_latitude = 45;

    do
    {
        if(!fgets(tempstr, 256, fle))
        {
            if (wgen[0].num[season_spring] > 0 && wgen[0].num[season_summer] > 0 &&
                wgen[0].num[season_autumn] > 0 && wgen[0].num[season_winter] > 0)
            {
                status = st_quit;
            }
            else
            {
                strcpy(tempstr, "Error: Incorrect format (");
                strcat(tempstr, filename);
                strcat(tempstr, ").");
                wgen_err(WGEN_ERR_INCORRECT_FILE, tempstr);
                wgen_exit();
                return -1;
            }
        }
        else
        {

        j = strlen(tempstr);

        line++;

/*-----------------------------------------------------------------------------
  | Check if we found the header for a new section                            |
  ----------------------------------------------------------------------------*/

        if (tempstr[0] == '[' && j > 3)
        {
            switch(tempstr[3])
            {
                case 'r':  /* [sp_r_ing] */

                            status = st_spring;
                            num_exceeded = 0;
                        break;
                case 'm':  /* [su_m_mer] */

                            status = st_summer;
                            num_exceeded = 0;
                        break;
                case 't':  /* [au_t_umn] */

                            status = st_autumn;
                            num_exceeded = 0;
                        break;
                case 'n':  /* [wi_n_ter] */

                            status = st_winter;
                            num_exceeded = 0;
                        break;

                case 'd':   /* [en_d_] */

                           status = st_quit;
                        break;

                case 'f':   /* [in_f_o] */

                            status = st_info;
                        break;

                case 'c':   /* [lo_c_ation] */

                            status = st_location;
                        break;
            }
        }

/*-----------------------------------------------------------------------------
  | Display any header found                                                  |
  ----------------------------------------------------------------------------*/

        else if (tempstr[0] != '/' && tempstr[0] != '#' && tempstr[0] != ';')
        {
            switch(status)
            {
            case st_info:
            case st_none:
                            wgen_info(tempstr);
                    break;

/*-----------------------------------------------------------------------------
  | ...or add to the weather entries                                          |
  ----------------------------------------------------------------------------*/

            case st_spring:
            case st_summer:
            case st_autumn:
            case st_winter:

                if (!num_exceeded)
                {
                    i = k = 0;

                    while(i < j)
                    {
                        switch(tempstr[i])
                        {
                            case '/':
                            case '(':
                            case ')':
                            case ' ':
                            case '\t':
                            case '\n':
                                    if (k > 0)
                                    {
                                        wparm[k] = '\0';

                                        if (!strcmp(wparm, "frzg") || !strcmp(wparm, "freezing"))
                                        {
                                            if (wgen[0].season[status][wgen[0].num[status]].temperature)
                                                wgen[0].season[status][wgen[0].num[status]].temperature |= tse_freezing;
                                            else
                                                wgen[0].season[status][wgen[0].num[status]].temperature = tpr_freezing;
                                        }
                                        else if (!strcmp(wparm, "cold"))
                                        {
                                            if (wgen[0].season[status][wgen[0].num[status]].temperature)
                                                wgen[0].season[status][wgen[0].num[status]].temperature |= tse_cold;
                                            else
                                                wgen[0].season[status][wgen[0].num[status]].temperature = tpr_cold;
                                        }
                                        else if (!strcmp(wparm, "cool"))
                                        {
                                            if (wgen[0].season[status][wgen[0].num[status]].temperature)
                                                wgen[0].season[status][wgen[0].num[status]].temperature |= tse_cool;
                                            else
                                                wgen[0].season[status][wgen[0].num[status]].temperature = tpr_cool;
                                        }
                                        else if (!strcmp(wparm, "warm"))
                                        {
                                            if (wgen[0].season[status][wgen[0].num[status]].temperature)
                                                wgen[0].season[status][wgen[0].num[status]].temperature |= tse_warm;
                                            else
                                                wgen[0].season[status][wgen[0].num[status]].temperature = tpr_warm;
                                        }
                                        else if (!strcmp(wparm, "hot"))
                                        {
                                            if (wgen[0].season[status][wgen[0].num[status]].temperature)
                                                wgen[0].season[status][wgen[0].num[status]].temperature |= tse_hot;
                                            else
                                                wgen[0].season[status][wgen[0].num[status]].temperature = tpr_hot;
                                        }
                                        else if (!strcmp(wparm, "rain"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_rain;
                                        else if (!strcmp(wparm, "snow"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_snow;
                                        else if (!strcmp(wparm, "hail"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_hail;
                                        else if (!strcmp(wparm, "~thunder"))
                                        {
                                                /* Mark 'uncertain' thunder with the snow-flag */

                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_thunder;
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_snow;
                                        }
                                        else if (!strcmp(wparm, "thunder"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_thunder;
                                        else if (!strcmp(wparm, "fog") || !strcmp(wparm, "mist"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_mist;
                                        else if (!strcmp(wparm, "heavy") || !strcmp(wparm, "steady"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_heavy;
                                        else if (!strcmp(wparm, "cloudy"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_cloudy;
                                        else if (!strcmp(wparm, "overcast"))
                                            wgen[0].season[status][wgen[0].num[status]].weather |= w_overcast;
                                        else switch(wparm[0])
                                        {
                                            case '0':
                                                    wgen[0].season[status][wgen[0].num[status]].wind_force = 0;
                                                break;
                                            case '1':
                                                    wgen[0].season[status][wgen[0].num[status]].wind_force = 1;
                                                break;
                                            case '2':
                                                    wgen[0].season[status][wgen[0].num[status]].wind_force = 2;
                                                break;
                                            case 'n':
                                            case 's':
                                                    wgen[0].season[status][wgen[0].num[status]].wind_direction = wparm[0] - ('a' - 'A');

                                                    if (k > 1)
                                                    {
                                                        switch(wparm[1])
                                                        {
                                                            case 'w':
                                                                wgen[0].season[status][wgen[0].num[status]].wind_direction += 1;
                                                               break;
                                                            case 'e':
                                                                wgen[0].season[status][wgen[0].num[status]].wind_direction += 2;
                                                               break;
                                                        }
                                                    }
                                                break;
                                            case 'e':
                                            case 'w':
                                                    wgen[0].season[status][wgen[0].num[status]].wind_direction = wparm[0] - ('a' - 'A');
                                                break;
                                        }
                                        k = 0;
                                    }
                                    break;
                            default:
                                    if (k < 9)
                                    {
                                        wparm[k] = tempstr[i];

                                        if (wparm[k] >= 'A' && wparm[k] <= 'Z')
                                            wparm[k] += ('a' - 'A');

                                        k++;
                                    }
                        }
                        i++;
                    }
                    if (wgen[0].num[status] < max_weather)
                    {
                        wgen[0].num[status]++;

                        if (wgen[0].num[status] == max_weather)
                            num_exceeded = 1;
                    }
                }
                else if (num_exceeded == 1)
                {
                    sprintf(tempstr, "Warning: Maximum number of entries for season #%i exceeded at line %i.", (status + 1), line);
                    wgen_err(WGEN_WARN_NUM_WEATHER, tempstr);
                    num_exceeded = 2;
                    break;
                }
                break;

                case st_location:
                case st_latitude:

                    i = k = 0;

                    while(i < j)
                    {
                        switch(tempstr[i])
                        {
                            case ' ':
                            case '=':
                            case ':':
                            case '\t':
                            case '\n':
                                    if (k > 0)
                                    {
                                        wparm[k] = '\0';

                                        if (status == st_location)
                                        {
                                            if (!strcmp(wparm, "latitude"))
                                                status = st_latitude;
                                        }
                                        else
                                        {
                                            status = st_location;
                                            wg_latitude = atoi(wparm);
                                        }
                                        k = 0;
                                    }
                            default:
                                    if (k < 9)
                                    {
                                        wparm[k] = tempstr[i];

                                        if (wparm[k] >= 'A' && wparm[k] <= 'Z')
                                            wparm[k] += ('a' - 'A');

                                        k++;
                                    }
                        }
                        i++;
                    }
                    status = st_location;
            }
        }
    }
    }
    while (status != st_quit);

/*-----------------------------------------------------------------------------
  | Close file and return                                                     |
  ----------------------------------------------------------------------------*/

    fclose(fle);

    return 0;
};

/*****************************************************************************
  * wgen_exit: Unload the weather generator                                  *
  ****************************************************************************/

void wgen_exit()
{
    free(weatherfile);
    free(wgen);
    wgen = NULL;
    weatherfile = NULL;
};
