/*
 *  A_WGEN Random Weather Generator     Copyright (C) Kimmo Kulovesi 1998
 *                                                    arkhan@softhome.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Please see 'license.txt'.
 *
 *****************************************************************************/

#ifndef arkhan_a_wgen_h
#define arkhan_a_wgen_h

/*****************************************************************************
  * The functions from 'a_wgen.c'                                            *
  ****************************************************************************/

typedef struct Weather
{
    unsigned char temperature,
                  precipitation,
                  cloud_cover,
                  wind_force;
    char          wind_direction;
}
Weather;

    /* Loads the weatherfile specified */

int wgen_load(char *filename);

    /* Generates a weather entry for the season specified. If the watch
       in question occurs (more than half of it, anyhow) at night, set
       'is_at_night'.

       The previous runs precipitation value should be passed as the last
       parameter. If not available, use '0'.
     */

Weather wgen_generate(unsigned char c_season, unsigned char is_at_night, unsigned char prev_precipitation);

    /* Unloads the weather generator */

void wgen_exit();

/*****************************************************************************
  * Variables from 'a_wgen.c' that are globally visible                      *
  ****************************************************************************/

extern signed char wgen_position;
extern char        *weatherfile;
extern int         wg_latitude;

/*****************************************************************************
  * Return values of the structure from 'wgen_generate'                      *
  ****************************************************************************/

    /* Temperatures */

#define temperature_freezing        0
#define temperature_cold            1
#define temperature_cool            2
#define temperature_warm            3
#define temperature_hot             4

    /* Precipitation (wet ones at the top end) */

#define p_none          0
#define p_mist          1
#define p_fog           2
#define p_snow          3
#define p_heavy_snow    4
#define p_blizzard      5
#define p_sleet         6
#define p_hail          7
#define p_heavy_sleet   8
#define p_rain          9
#define p_heavy_rain    10
#define p_thunder       11
#define p_thunder_hail  12

    /* Percentage covered. Note that the actual return value is anything *
     * within the scale (clear = 0-19, cloudy 20-79, overcast 80-100).   */

#define sky_clear                   0
#define sky_cloudy                  20
#define sky_overcast                80

    /* Wind force (sheltered areas should adjust this down, and open waters *
     * probably up).                                                        */

#define wind_calm                   0
#define wind_breeze                 1
#define wind_windy                  2
#define wind_gale                   3
#define wind_storm                  4

    /* Wind directions (derived from ASCII values). */

#define dir_N                       ('N')
#define dir_S                       ('S')
#define dir_NW                      ('N' + 1)
#define dir_NE                      ('N' + 2)
#define dir_SW                      ('S' + 1)
#define dir_SE                      ('S' + 2)

    /* These are not actually used (hexes have no direct east/west), but are *
     * included for compatibility.                                           */

#define dir_E                       ('E')
#define dir_W                       ('W')

/*****************************************************************************
  * Needed by WGEN, but not in 'a_wgen.c'                                    *
  ****************************************************************************/

    /* Called by wgen to provide error messages */

void wgen_err(unsigned char num, char *wgen_err);

    /* Called by wgen to display information headers from files */

void wgen_info(char *wgen_inf);

/*****************************************************************************
  * Internal defines (using a different structure to save memory)            *
  ****************************************************************************/

    /* Weather flags (precipitation/cloud cover/etc) */

#define w_rain          1
#define w_snow          2
#define w_heavy         4
#define w_thunder       8
#define w_cloudy        16
#define w_overcast      32
#define w_hail          64
#define w_mist          128

    /* Temperatures */

#define tpr_freezing   0
#define tpr_cold       1
#define tpr_cool       2
#define tpr_warm       3
#define tpr_hot        4

    /* Temperatures at night, if different than at daytime */

#define tse_hot        8
#define tse_warm       16
#define tse_cool       32
#define tse_cold       64
#define tse_freezing   128

    /* wgen_prev flags */

#define i_thunder       1
#define i_mist          2
#define i_rain          4
#define i_heavy_rain    8
#define i_snow          16
#define i_heavy_snow    32

    /* Maximum number of weather entries / season */

#define max_weather    20

    /* Error codes (sent to 'wgen_err') */

#define WGEN_WARN_NUM_WEATHER       1   /* Too many entries in file     */
#define WGEN_ERR_NOT_READY          10  /* Not loaded when tried to run */
#define WGEN_ERR_INCORRECT_FILE     20  /* Incorrect file structure     */
#define WGEN_ERR_NO_MEMORY_LOAD     22  /* Not enough memory to load    */
#define WGEN_ERR_NO_FILE            25  /* File not found               */

    /* Version number  (a.b) */

#define wg_ver_a    1
#define wg_ver_b    1

#endif
