/*
 *  A_WORLD                             Copyright (C) Kimmo Kulovesi 1998
 *                                                    arkhan@softhome.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Please see 'license.txt'.
 *
 *****************************************************************************/

#include <string.h>
#include "a_world.h"

/*****************************************************************************
 * x_get_month_name: Get the name of each month                              *
 *****************************************************************************/

char *harn_get_month_name(signed char month)
{
    static char abuf[15];

    switch(month)
    {
        case m_nuzyael:     strcpy(abuf, "Nuzyael");    break;
        case m_peonu:       strcpy(abuf, "Peonu");      break;
        case m_kelen:       strcpy(abuf, "Kelen");      break;
        case m_nolus:       strcpy(abuf, "Nolus");      break;
        case m_larane:      strcpy(abuf, "Larane");     break;
        case m_agrazhar:    strcpy(abuf, "Agrazhar");   break;
        case m_azura:       strcpy(abuf, "Azura");      break;
        case m_halane:      strcpy(abuf, "Halane");     break;
        case m_savor:       strcpy(abuf, "Savor");      break;
        case m_ilvin:       strcpy(abuf, "Ilvin");      break;
        case m_navek:       strcpy(abuf, "Navek");      break;
        case m_morgat:      strcpy(abuf, "Morgat");     break;
        default:            strcpy(abuf, "");
    }
    return abuf;
};

char *earth_get_month_name(signed char month)
{
    static char abuf[15];

    switch(month)
    {
        case m_january:     strcpy(abuf, "January");    break;
        case m_february:    strcpy(abuf, "February");   break;
        case m_march:       strcpy(abuf, "March");      break;
        case m_april:       strcpy(abuf, "April");      break;
        case m_may:         strcpy(abuf, "May");        break;
        case m_june:        strcpy(abuf, "June");       break;
        case m_july:        strcpy(abuf, "July");       break;
        case m_august:      strcpy(abuf, "August");     break;
        case m_september:   strcpy(abuf, "September");  break;
        case m_october:     strcpy(abuf, "October");    break;
        case m_november:    strcpy(abuf, "November");   break;
        case m_december:    strcpy(abuf, "December");   break;
        default:            strcpy(abuf, "");
    }
    return abuf;
};

/*****************************************************************************
 * x_get_year_sign: Get the year suffix ('AD', 'BC', 'TR', 'BT').            *
 *****************************************************************************/

char *harn_get_year_sign(signed long year)
{
    static char ybuf[3];

    if (year < 0)
        strcpy(ybuf, "BT");
    else
        strcpy(ybuf, "TR");

    return ybuf;
};

char *earth_get_year_sign(signed long year)
{
    static char ybuf[3];

    if (year < 0)
        strcpy(ybuf, "BC");
    else
        strcpy(ybuf, "AD");

    return ybuf;
};

/*****************************************************************************
 * x_get_season: Return the season of the given month and day                *
 *****************************************************************************/

unsigned char harn_get_season(signed char month, signed char day)
{
    switch(month)
    {
        case m_nuzyael:
        case m_peonu:
        case m_kelen:
                    return season_spring;
        case m_nolus:
        case m_larane:
        case m_agrazhar:
                    return season_summer;
        case m_azura:
        case m_halane:
        case m_savor:
                    return season_autumn;
        default:
                        /* Ilvin, Navek and Morgat */

                    return season_winter;
    }
};

unsigned char earth_get_season(signed char month, signed char day)
{
        /* Returns the seasons on earth, according to the approximate
         * climate of the UK.
         */

    switch(month)
    {
            /* Winter from the middle of November to the middle of February */

        case m_november:
                if (day < 15)
                    return season_autumn;
        case m_december:
        case m_january:
                return season_winter;

            /* Spring from the middle of February to the middle of May */

        case m_february:
                if (day < 15)
                    return season_winter;
        case m_march:
        case m_april:
                return season_spring;

            /* Summer from the middle of May to the middle of August */

        case m_may:
                if (day < 17)
                    return season_spring;
        case m_june:
        case m_july:

            /* Autumn from the middle of August to the middle of November */

        case m_august:
                if (day < 16)
                    return season_summer;

        default:

            /* September and October */

                return season_autumn;
    }
};
/*****************************************************************************
 * x_num_of_days: Return the number of days at a give month and year         *
 *****************************************************************************/

int harn_num_of_days(signed char month, signed long year)
{
    if (month > 0)
        return 30;
    else
        return 360;
};

unsigned char is_leap_year(signed long year)
{
    if (year == 0)  /* Avoid division by zero, as well as correct the year */
        year = 1;

    return (!(year % 4) && (year % 100) || !(year % 400));
};

int earth_num_of_days(signed char month, signed long year)
{
    switch(month)
    {
        case 0:
                        return (365 + is_leap_year(year));
        case m_april:
        case m_june:
        case m_september:
        case m_november:
                        return 30;
        case m_february:
                        return (28 + is_leap_year(year));
    }
    return 31;
};

/*****************************************************************************
 * x_day_of_the_year: Convert a date to the day of the year (dd,mm,yyyy->ddd)*
 *****************************************************************************/

int harn_day_of_the_year(signed char day, signed char month, signed long year)
{
    return (((month - 1) * 30) + day);
};

int earth_day_of_the_year(signed char day, signed char month, signed long year)
{
    int doy = 0;

    month--;
    while(month)
        doy += earth_num_of_days(month--, year);

    return (doy + day);
};

/*****************************************************************************
 * earth_weekday: Get the name of the day of the week at a given date        *
 *****************************************************************************/

char *earth_weekday(signed char day, signed char month, signed long year)
{
    static char dbuf[10];
    int i;

    i = year + (month - 14) / 12;

    i = (   (   (
                    13 *
                    (
                        month + 10
                        - (month + 10)
                        / 13
                        * 12
                    )
                    - 1
                )
                / 5
                + day
                + 77
                + 5
                * (i % 100)
                / 4
                + i / 400
                - i / 100
                * 2
            )
            % 7
        );

    switch(i)
    {
        case 0:     strcpy(dbuf, "Sunday");     break;
        case 1:     strcpy(dbuf, "Monday");     break;
        case 2:     strcpy(dbuf, "Tuesday");    break;
        case 3:     strcpy(dbuf, "Wednesday");  break;
        case 4:     strcpy(dbuf, "Thursday");   break;
        case 5:     strcpy(dbuf, "Friday");     break;
        case 6:     strcpy(dbuf, "Saturday");
    }

    return dbuf;
};
