/*
 *  Arkhan's Random Weather Generator   Copyright (C) Kimmo Kulovesi 1998
 *                                                    arkhan@softhome.net
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Please see 'license.txt'.
 *
 *****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <math.h>
#include "a_dice.h"
#include "a_world.h"
#include "a_wgen.h"
#include "astral.h"
#include "wgen.h"

unsigned char   current_day = 1,
                current_month = 1;
signed long     current_year = 0;
signed char     wg_moon_age = -1;

unsigned char   wg_flags = 0;

/*****************************************************************************
  * The main function                                                        *
  ****************************************************************************/

void ascii_print_newday() { return; };

int main(int argc, char *argv[])
{
    int             j = 0;
    signed char     i,
                    k,
                    l,
                    m,
                    n,
                    day_spacing = -1;
    char            filename[128],
                    dbuf[12],
                    *monthname;
    Weather         w[7];
    long            num_to_generate = -999;
    FILE            *fle;

    unsigned char (*num_of_days)(signed char, signed long);
    unsigned char (*get_season)(signed char, signed long);
    void (*print_newday)();
    void (*print_date)(signed char, char *);
    void (*print_weekday)();
    void (*print_sun)(signed char, signed char, signed char, signed char);
    void (*print_moon)();
    void (*print_newline)();
    void (*print_temperature)(unsigned char);
    void (*print_weather)(unsigned char);
    void (*print_sky)(unsigned char);
    void (*print_wind)(char, unsigned char);
    void (*print_newmonth)(char *);

/*----------------------------------------------------------------------------
  | Display program usage and quit if not enough parameters                  |
  ---------------------------------------------------------------------------*/

    if (argc < 2)
    {
        wg_usage();
        return 0;
    }

/*----------------------------------------------------------------------------
  | Check command line parameters for switches                               |
  ---------------------------------------------------------------------------*/

    filename[0] = '\0';
    wg_latitude = -500;

        /* Initialize random number generator */

    Init_Dice(time(NULL));

    for (i = argc - 1; i; i--)
    {
            /* Check if it is an option switch */

        if (argv[i][0] == '-')
        {
            n = 0;
            j = 1;
            m = strlen(argv[i]);
            k = -1;

            while (j <= m)
            {
                if (k >= 0 || j == m)
                {
                    if (j < m && ((argv[i][j] >= '0' && argv[i][j] <= '9') || argv[i][j] == '-' || argv[i][j] == '+'))
                    {
                        dbuf[k++] = argv[i][j];
                    }
                    else
                    {
                        if (k > 0)
                        {
                            dbuf[k] = '\0';

                            switch(n)
                            {
                                case wg_fl_custom_lat:
                                                    wg_latitude = atoi(dbuf);
                                                    n = 0;
                                                break;
                                case wg_fl_custom_breaks:
                                                    day_spacing = atoi(dbuf);
                                                    n = 0;
                                                break;
                            }
                        }
                        if (j < m)
                        {
                            k = -1;
                            j--;
                        }
                    }
                }
                else switch(argv[i][j])
                {
                    case 'c':
                    case 'C':
                                wg_flags |= wg_fl_run_continue;
                        break;
                    case 's':
                    case 'S':
                                wg_flags |= wg_fl_save_continue;
                        break;
                    case 'p':
                    case 'P':
                                if (wg_flags & wg_fl_moon_age)
                                    fprintf(stderr, "Warning: Display of both moon age and percents illuminated not supported.\n");
                                else
                                    wg_flags |= wg_fl_moon_percent;
                        break;
                    case 'a':
                    case 'A':
                                if (wg_flags & wg_fl_moon_percent)
                                    fprintf(stderr, "Warning: Display of both moon age and percents illuminated not supported.\n");
                                else
                                    wg_flags |= wg_fl_moon_age;
                        break;
                    case 'l':
                    case 'L':
                                n = wg_fl_custom_lat;
                                k = 0;
                        break;
                    case 'b':
                    case 'B':
                                n = wg_fl_custom_breaks;
                                k = 0;
                        break;
                    case 'h':
                    case 'H':
                                wg_flags |= wg_fl_html;
                        break;
                    case 'e':
                    case 'E':
                                wg_flags |= wg_fl_earth;
                        break;
                    default:
                                fprintf(stderr, "Warning: Incorrect option switch \"%c\" ignored.\n", argv[i][j]);
                }
                j++;
            }
        }

/*----------------------------------------------------------------------------
  | See if the parameter is a number                                         |
  ---------------------------------------------------------------------------*/

        else if (j > -666 && argc > 2 && ((argv[i][0] >= '0' && argv[i][0] <= '9') || argv[i][0] == '+'))
        {
            j = strlen(argv[i]) - 1;
            k = m = 0;

            while(j > 0)
            {
               /*  Check for any non-numeric characters, as it might be an
                *  odd filename, or the date
                */

                if (argv[i][j] < '0' || argv[i][j] > '9')
                {
                    if (argv[i][j] == '.' || argv[i][j] == '/')
                        k++;
                    else if (k == 0)
                    {
                        switch(argv[i][j])
                        {
                            case 'm':
                            case 'M':
                                        if (j == strlen(argv[i]) - 1)
                                        {
                                            wg_flags |= wg_fl_do_months;
                                        }
                                        else
                                        {
                                            j = -666;
                                            i++;
                                        }
                                    break;
                            case '-':
                                        if (k == 0)
                                            break;
                            default:
                                        j = -666;
                                        i++;
                        }
                    }
                }
                j--;
            }

            if (j > -666)
            {
                if (k > 0)
                {
                    if (k > 2)
                    {
                        j = -666;
                        i++;
                    }
                    else
                    {
                        k = strlen(argv[i]);
                        j = l = m = 0;

                        do
                        {
                            if (argv[i][j] == '.' || argv[i][j] == '/' || j == k)
                            {
                                if ((l > 0 && l < 3) || m > 1)
                                {
                                    dbuf[l] = '\0';
                                    l = 0;

                                    if (m == 0)
                                        current_day = atoi(dbuf);
                                    else if (m == 1)
                                        current_month = atoi(dbuf);
                                    else
                                    {
                                        current_year = atol(dbuf);

                                        if (current_year == 0)
                                            current_year = 1;
                                    }
                                    m++;
                                }
                                else
                                {
                                    j = -666;
                                    i++;
                                    break;
                                }
                            }
                            else
                            {
                                dbuf[l] = argv[i][j];
                                l++;
                            }

                            j++;
                        }
                        while(j <= k);
                    }
                }
                else if (num_to_generate != -999)
                {
                    j = -666;
                    i++;
                }
                else
                {
                    if (wg_flags & wg_fl_do_months)
                        argv[i][strlen(argv[i]) - 1] = '\0';

                    num_to_generate = (long)atol(argv[i]);
                }
            }
        }

/*----------------------------------------------------------------------------
  | If not, then it must be our filename                                     |
  ---------------------------------------------------------------------------*/

        else
        {
            j = 0;

            if (filename[0] != '\0')
                fprintf(stderr, "Warning: Multiple file names ignored: \"%s\".\n", filename);

            strcpy(filename, argv[i]);
        }
    }
/*----------------------------------------------------------------------------
  | Check that the parameters are properly set                               |
  ---------------------------------------------------------------------------*/

    if (wg_flags & wg_fl_run_continue)
    {
        fle = fopen("wgen.cnt", "rb");

        if (fle == NULL)
        {
            fprintf(stderr, "Error: Cannot load data from previous generation, aborting.\n");
            return 5;
        }

        j = wg_flags;

        fread       ((&wgen_position),    sizeof(char),       1, fle);
        fread       ((&current_day),      sizeof(char),       1, fle);
        fread       ((&current_month),    sizeof(char),       1, fle);
        fread       ((&current_year),     sizeof(long),       1, fle);
        fread       ((&wg_latitude),      sizeof(int),        1, fle);
        fread       ((&wg_flags),         sizeof(char),       1, fle);
        fread       ((&day_spacing),      sizeof(char),       1, fle);
        fread       ((&w[6]),             sizeof(Weather),    1, fle);
        fread       ((&filename),         sizeof(char),     128, fle);
        fclose(fle);

        if (j & wg_fl_save_continue)
            wg_flags |= wg_fl_save_continue;
        else if (wg_flags & wg_fl_save_continue)
            wg_flags -= wg_fl_save_continue;

        if (!(j & wg_fl_do_months))
            if (wg_flags & wg_fl_do_months)
                wg_flags -= wg_fl_do_months;

        wg_flags |= wg_fl_run_continue;
    }

    if (filename[0] == '\0')    /* Show usage and quit, if no filename */
    {
        if (num_to_generate != -999)
        {
            sprintf(filename, "%i\0", num_to_generate);
            num_to_generate = -999;
        }
        else
        {
            wg_usage();
            return 1;
        }
    }

    if (current_year == 0)
    {
        if (wg_flags & wg_fl_earth)
            current_year = 1190;
        else
            current_year = 720;
    }

/*----------------------------------------------------------------------------
  | Prepare for generation                                                   |
  ---------------------------------------------------------------------------*/

    wg_header(wg_header_title);

    if (wg_latitude != -500)
        j = wg_latitude;
    else
        j = -500;

    if (wgen_load(filename))
        return 2;

    if (j != -500)
        wg_latitude = j;

    if (wg_flags & wg_fl_earth)
    {
        num_of_days     = (e_num_of_days);
        get_season      = (e_get_season);
    }
    else
    {
        num_of_days     = (h_num_of_days);
        get_season      = (h_get_season);
    }

    if (current_day > num_of_days(current_month, current_year))
    {
        current_day = 1;
        current_month++;

        while(current_month > num_of_months)
        {
            current_month -= num_of_months;
            current_year++;
        }
    }

    if (wg_flags & wg_fl_earth)
        monthname = earth_get_month_name(current_month);
    else
        monthname = harn_get_month_name(current_month);

    if (num_to_generate < 1)
    {
        wg_flags |= wg_fl_do_months;
        num_to_generate = 1;
    }

    wg_header(wg_header_begin);

    if (wg_flags & wg_fl_html)
    {
        print_newday      = (html_print_newday);
        print_date        = (html_print_date);
        print_weekday     = (html_print_weekday);
        print_sun         = (html_print_sun);
        print_moon        = (html_print_moon);
        print_newline     = (html_print_newline);
        print_temperature = (html_print_temperature);
        print_sky         = (html_print_sky);
        print_wind        = (html_print_wind);
        print_weather     = (html_print_weather);
        print_newmonth    = (html_print_newmonth);
    }
    else
    {
        print_newday      = (ascii_print_newday);
        print_date        = (ascii_print_date);
        print_weekday     = (ascii_print_weekday);
        print_sun         = (ascii_print_sun);
        print_moon        = (ascii_print_moon);
        print_newline     = (ascii_print_newline);
        print_temperature = (ascii_print_temperature);
        print_sky         = (ascii_print_sky);
        print_wind        = (ascii_print_wind);
        print_weather     = (ascii_print_weather);
        print_newmonth    = (ascii_print_newmonth);
    }

    if (!(wg_flags & wg_fl_run_continue))
    {
        /* Generate the last watch of previous day for internal reference */

        if (current_day > 1)
            w[6] = wgen_generate(get_season(current_month, (current_day - 1)), 1, 0);
        else if (current_month > 1)
            w[6] = wgen_generate(get_season((current_month - 1), num_of_days((current_month - 1), current_year)), 1, 0);
        else
            w[6] = wgen_generate(get_season(num_of_months, num_of_days(num_of_months, (current_year - 1))), 1, 0);
    }

    if (wg_flags & wg_fl_moon_age)
    {
        if (wg_flags & wg_fl_earth)
            wg_moon_age = earth_age_of_moon(current_day, current_month, current_year);
        else
            wg_moon_age = harn_age_of_moon(current_day, current_month, current_year);

        wg_moon_age--;
    }

/*----------------------------------------------------------------------------
  | Enter the actual generation loop                                         |
  ---------------------------------------------------------------------------*/

    do
    {
        w[0] = w[6];

        j = get_season(current_month, current_day);

        sun_times(&k, &l, &m, &n);

        for (i = 1; i <= 6; i++)
            w[i] = wgen_generate(j, is_night_watch(i, k, l, m, n), w[i - 1].precipitation);

        print_newday();
        print_date(current_day, monthname);
        for (i = 6; i; i--)
            print_temperature(w[i].temperature);
        print_newline();

        print_weekday();
        for (i = 6; i; i--)
            print_sky(w[i].cloud_cover);
        print_newline();

        print_sun(k, l, m, n);
        for (i = 6; i; i--)
            print_wind(w[i].wind_direction, w[i].wind_force);
        print_newline();

        print_moon(0);
        for (i = 6; i; i--)
            print_weather(w[i].precipitation);

            /* Print day breaks */

        if (day_spacing >= 0)
        {
            i = day_spacing;
            if (wg_flags & wg_fl_html)
            {
                printf("<TR><TD colspan=\"9\">\n");
                if (i > 0)
                    printf("    &nbsp;");

                while (i > 1)
                {
                    printf("\n<BR>&nbsp;");
                    i--;
                }
                printf("\n</TD></TR>\n");
            }
            else while (i > 0)
            {
                printf("\n");
                i--;
            }
        }

        if (wg_flags & wg_fl_html)
            printf("</TABLE><BR>\n");
        else
            printf("\n\n");

        current_day++;

        if (!(wg_flags & wg_fl_do_months))
            num_to_generate--;

        if (current_day > num_of_days(current_month, current_year))
        {
            current_day = 1;
            current_month++;

            if (wg_flags & wg_fl_do_months)
                num_to_generate--;

            if (current_month > num_of_months)
            {
                current_month = 1;
                current_year++;

                if (current_year == 0)
                    current_year = 1;

                if (num_to_generate > 0)
                    wg_header(wg_header_year);
            }

            if (num_to_generate > 0)
            {
                if (wg_flags & wg_fl_earth)
                    monthname = earth_get_month_name(current_month);
                else
                    monthname = harn_get_month_name(current_month);

                if (current_month != 1)
                    print_newmonth(monthname);
            }
        }
    }
    while(num_to_generate);

    if (wg_flags & wg_fl_save_continue)
    {
        fle = fopen("wgen.cnt", "wb");

        if (fle == NULL)
            fprintf(stderr, "Error: Could not save data for continued generation.\n");
        else
        {
            fwrite       ((&wgen_position),    sizeof(char),       1, fle);
            fwrite       ((&current_day),      sizeof(char),       1, fle);
            fwrite       ((&current_month),    sizeof(char),       1, fle);
            fwrite       ((&current_year),     sizeof(long),       1, fle);
            fwrite       ((&wg_latitude),      sizeof(int),        1, fle);
            fwrite       ((&wg_flags),         sizeof(char),       1, fle);
            fwrite       ((&day_spacing),      sizeof(char),       1, fle);
            fwrite       ((&w[6]),             sizeof(Weather),    1, fle);
            fwrite       ((&filename),         sizeof(char),     128, fle);
            fclose(fle);
        }
    }

/*----------------------------------------------------------------------------
  | Exit the program                                                         |
  ---------------------------------------------------------------------------*/

    wg_header(wg_header_exit);
    wgen_exit();
    return 0;
};

/*****************************************************************************
  * wg_header: Prints the headers (and footers) used by the generator        *
  ****************************************************************************/

void wg_header(unsigned char mode)
{
    char abuf[3];

    if (mode == wg_header_year || mode == wg_header_begin)
    {
        if (wg_flags & wg_fl_earth)
            strcpy(abuf, earth_get_year_sign(current_year));
        else
            strcpy(abuf, harn_get_year_sign(current_year));
    }

    if (wg_flags & wg_fl_html)
    {
        switch(mode)
        {
            case wg_header_begin:
                    if (wg_latitude < 0)
                        printf("<B>%i&deg; South</B>", -wg_latitude);
                    else if (wg_latitude == 0)
                        printf("<B>0&deg;</B>");
                    else
                        printf("<B>%i&deg; North</B>", wg_latitude);
            case wg_header_year:

                    if (current_year < 0)
                        printf("\n<H2>%li %s", ((long)(-current_year)), abuf);
                    else
                        printf("\n<H2>%li %s", ((long)current_year), abuf);

                    if (wg_flags & wg_fl_earth)
                        if (is_leap_year(current_year))
                            printf(" (Leap year)");

                    printf(" <HR width\"100%%\"></H2>\n");
                return;

            case wg_header_title:
                    printf("<HTML>\n");
                    printf("<HEAD>\n");
                    printf("    <TITLE>Arkhan's Random Weather Generator v%i.%i</TITLE>\n", wg_ver_a, wg_ver_b);
                    printf("    <META name=\"Description\" content=\"WGEN (C) Kimmo Kulovesi 1998\">\n");
                    printf("</HEAD>\n");
                    printf("<BODY>\n");
                return;

            case wg_header_exit:
                    printf("</BODY>\n</HTML>\n");
                return;
        }
    }
    else
    {
        switch(mode)
        {
            case wg_header_begin:
                        printf("\n");
                        if (wg_latitude < 0)
                            printf("%i degrees South\n", -wg_latitude);
                        else if (wg_latitude == 0)
                            printf("0 degrees\n");
                        else
                            printf("%i degrees North\n", wg_latitude);
            case wg_header_year:
                    if (current_year < 0)
                       printf("\n    %li %s", ((long)(-current_year)), abuf);
                    else
                       printf("\n    %li %s", ((long)current_year), abuf);

                    if (wg_flags & wg_fl_earth)
                        if (is_leap_year(current_year))
                            printf(" (Leap year)");

                    printf("\n-------------------------------------------------------------------------------\n");
            case wg_header_title:
                return;

            case wg_header_exit:
                    printf("-------------------------------------------------------------------------------\n");
                    printf("Arkhan's Random Weather Generator v%1i.%-2i      Copyright (C) Kimmo Kulovesi 1998\n", wg_ver_a, wg_ver_b);
                return;
        }
    }
};

/*****************************************************************************
  | wg_usage: Prints information about the command line parameters and quits |
  ****************************************************************************/

void wg_usage()
{
    printf("Arkhan's Random Weather Generator v%1i.%-2i      Copyright (C) Kimmo Kulovesi 1998\n", wg_ver_a, wg_ver_b);
    printf("                                                           arkhan@softhome.net\n\n");

    printf("Usage: wgen <datafile> [##] [dd.mm.yyyy] [-h -e -l## -b## -p -a -s -c]\n\n");

    printf("            [##]         - Amount to generate; either a number of days\n");
    printf("                           or a number of months followed by an 'M'.\n\n");

    printf("            [dd.mm.yyyy] - The date to start at (the year is optional).\n\n");

    printf("            -e           - Generate as for Earth.\n");
    printf("            -h           - Encode output in HTML.\n");
    printf("            -l##         - Set latitude to ## degrees (negative = southern).\n");
    printf("            -b##         - Break days ## lines apart.\n");
    printf("            -p           - Display moon phase as percents illuminated.\n");
    printf("            -a           - Display the age of moon in days.\n\n");
    printf("            -s           - Save data for continuing this run later.\n");
    printf("            -c           - Continue generation from saved data.\n\n");

    printf("    This program is free software under the terms of the GNU General Public\n");
    printf("    License. Distributed with ABSOLUTELY NO WARRANTY; use at your own risk.\n");
};

/*****************************************************************************
  * wgen_info: Displays weatherfile headers                                  *
  ****************************************************************************/

void wgen_info(char *wgen_inf)
{
    int i;
    char *j;
    char tempstr[256];

    if (wgen_inf == NULL)
        return;

    if (wgen_inf[0] == '*' || wgen_inf[0] == '@')
    {
        if (wg_flags & wg_fl_html)
        {
            if (wgen_inf[0] == '*')
                return;
        }
        else if (wgen_inf[0] == '@')
                return;

        i = strlen(wgen_inf) - 1;

        if (i > 0)
        {
            i--;
            tempstr[i] = '\0';

            while(i > 0)
            {
                tempstr[i - 1] = wgen_inf[i];
                i--;
            }
        }
    }
    else
    {
        strcpy(tempstr, wgen_inf);
        i = strlen(tempstr) - 1;

        tempstr[i] = '\0';
    }

    if (wg_flags & wg_fl_html)
        printf("<B> %s </B>\n<BR>", tempstr);
    else
        printf("%s", tempstr);
};

/*****************************************************************************
  * wgen_err: Displays error messages                                        *
  ****************************************************************************/

void wgen_err(unsigned char num, char *wgen_err)
{
    fprintf(stderr, "%s\n", wgen_err);
};

/*****************************************************************************
  * Functions to print the weather entries (ASCII)                           *
  ****************************************************************************/

void ascii_print_wind(char d, unsigned char f)
{
    printf("  ");

    switch(d)
    {
        case dir_NW: printf("NW:");  break;
        case dir_NE: printf("NE:");  break;
        case dir_SW: printf("SW:");  break;
        case dir_SE: printf("SE:");  break;
        case dir_N:  printf("N: ");  break;
        case dir_E:  printf("E: ");  break;
        case dir_S:  printf("S: ");  break;
        default:     printf("W: ");  break;
    }

    switch(f)
    {
        case wind_calm:     printf("Calm ");    break;
        case wind_breeze:   printf("Light");    break;
        case wind_windy:    printf("Windy");    break;
        case wind_gale:     printf("Gale ");    break;
        case wind_storm:    printf("Storm");    break;
    }
};

void ascii_print_sky(unsigned char i)
{
    if (i >= sky_overcast)
        printf("  Overcast");
    else if (i >= sky_cloudy)
        printf("  Cloudy  ");
    else
        printf("  Clear   ");
};

void ascii_print_temperature(unsigned char i)
{
    switch(i)
    {
        case temperature_freezing:  printf("  Freezing"); break;
        case temperature_cold:      printf("  Cold    "); break;
        case temperature_cool:      printf("  Cool    "); break;
        case temperature_warm:      printf("  Warm    "); break;
        case temperature_hot:       printf("  Hot     ");
    }
};

void ascii_print_weather(unsigned char i)
{
    switch(i)
    {
        case p_fog:         printf("  Fog     ");     break;
        case p_mist:        printf("  Mist    ");     break;
        case p_rain:        printf("  Rain    ");     break;
        case p_sleet:       printf("  Sleet   ");     break;
        case p_snow:        printf("  Snow    ");     break;
        case p_heavy_rain:  printf("  Hv. Rain");     break;
        case p_heavy_sleet: printf("  Hv.Sleet");     break;
        case p_heavy_snow:  printf("  Hv. Snow");     break;
        case p_thunder:     printf("  Thunder ");     break;
        case p_thunder_hail:printf("  Thunder*");     break;
        case p_hail:        printf("  Hail    ");     break;
        case p_blizzard:    printf("  Blizzard");     break;
        default:            printf("          ");     break;
    }
};

void ascii_print_date(signed char current_day, char *c_month)
{
    printf("%2i. %-15s", current_day, c_month);
};

void ascii_print_sun(signed char rise_h, signed char rise_m, signed char set_h, signed char set_m)
{
    printf("Sun:  ");

    timep(rise_h);
    printf(":");
    timep(rise_m);

    printf(" - ");

    timep(set_h);
    printf(":");
    timep(set_m);
};

void ascii_print_moon()
{
    signed char i;

    if (wg_flags & wg_fl_earth)
    {
        i = earth_moon_phase(current_day, current_month, current_year);
        printf("Luna: ");
    }
    else
    {
        i = harn_moon_phase(current_day, current_month, current_year);
        printf("Yael: ");
    }

    if (i == 0)
        wg_moon_age = 0;
    else
        wg_moon_age++;

    switch(i)
    {
        case -100:  printf("Full Moon    ");  break;
        case 50:    printf("First Quarter");  break;
        case -50:   printf("Last Quarter ");  break;
        case 0:     printf("New Moon     ");  break;
        default:
            if (wg_flags & wg_fl_moon_percent)
            {
                if (i > 50)
                    printf("+Gibbous  %2i%%", i);
                else if (i > 0)
                    printf("+Crescent %2i%%", i);
                else if (i > -50)
                    printf("-Crescent %2i%%", -i);
                else
                    printf("-Gibbous  %2i%%", -i);
            }
            else if (wg_flags & wg_fl_moon_age)
            {
                if (i > 50)
                    printf("+Gibbous ");
                else if (i > 0)
                    printf("+Crescent");
                else if (i > -50)
                    printf("-Crescent");
                else
                    printf("-Gibbous ");

                printf(" %2id", wg_moon_age);
            }
            else
            {
                if (i > 0)
                    printf("Wax.");
                else
                {
                    i = -i;
                    printf("Wan.");
                }

                if (i > 50)
                    printf(" Gibbous ");
                else
                    printf(" Crescent");
            }
    }
};

void ascii_print_newline()
{
    printf("\n");
};

void ascii_print_weekday()
{
    if (wg_flags & wg_fl_earth)
        printf("%19s", earth_weekday(current_day, current_month, current_year));
    else
        printf("                   ");
};

void ascii_print_newmonth(char *c_month)
{
    printf("\n-------------------");

    if (!strcmp(c_month, "Kelen"))
        printf("  Kelen/Saraju\n\n\n");
    else
        printf("  %s\n\n\n", c_month);
};

/*****************************************************************************
  * Functions to print the weather entries (HTML)                            *
  ****************************************************************************/

void html_print_wind(char d, unsigned char f)
{
    printf("<TD>");

    switch(d)
    {
        case dir_NW: printf("NW: ");        break;
        case dir_NE: printf("NE: ");        break;
        case dir_SW: printf("SW: ");        break;
        case dir_SE: printf("SE: ");        break;
        case dir_N:  printf("N:&nbsp;");    break;
        case dir_E:  printf("E:&nbsp; ");   break;
        case dir_S:  printf("S:&nbsp; ");   break;
        default:     printf("W:&nbsp; ");   break;
    }

    switch(f)
    {
        case wind_calm:     printf("Calm");     break;
        case wind_breeze:   printf("Breeze");   break;
        case wind_windy:    printf("Windy");    break;
        case wind_gale:     printf("Gale");     break;
        case wind_storm:    printf("Storm");
    }
    printf("</TD>\n");
};

void html_print_sky(unsigned char i)
{
    printf("<TD>");

    if (i >= sky_overcast)
        printf("Overcast");
    else if (i >= sky_cloudy)
        printf("Cloudy");
    else
        printf("Clear");

    printf("</TD>\n");
};

void html_print_temperature(unsigned char i)
{
    printf("<TD width=\"13%%\">");

    switch(i)
    {
        case temperature_freezing:  printf("Freezing"); break;
        case temperature_cold:      printf("Cold");     break;
        case temperature_cool:      printf("Cool");     break;
        case temperature_warm:      printf("Warm");     break;
        case temperature_hot:       printf("Hot");
    }
    printf("</TD>\n");
};

void html_print_weather(unsigned char i)
{
    printf("<TD>");

    switch(i)
    {
        case p_mist:        printf("Mist");             break;
        case p_fog:         printf("Fog");              break;
        case p_rain:        printf("Rain");             break;
        case p_sleet:       printf("Sleet");            break;
        case p_snow:        printf("Snow");             break;
        case p_heavy_rain:  printf("Heavy Rain");       break;
        case p_heavy_sleet: printf("Heavy Sleet");      break;
        case p_heavy_snow:  printf("Heavy Snow");       break;
        case p_thunder:     printf("Thunder");          break;
        case p_thunder_hail:printf("Thunder/Hail");     break;
        case p_hail:        printf("Hail");             break;
        case p_blizzard:    printf("Blizzard");         break;
        default:            printf("&nbsp;");           break;
    }
    printf("</TD>\n");
};

void html_print_date(signed char current_day, char *c_month)
{
    printf("<TR>\n<TD width=\"5%%\"><TT><B>%i.</B></TT></TD>\n<TD width=\"16%%\"><TT><B>%s</B></TT></TD>\n<TD width=\"1%%\"></TD>\n", current_day, c_month);
};

void html_print_sun(signed char rise_h, signed char rise_m, signed char set_h, signed char set_m)
{
    printf("<TD>Sun:</TD><TD>");

    timep(rise_h);
    printf(":");
    timep(rise_m);

    printf(" - ");

    timep(set_h);
    printf(":");
    timep(set_m);

    printf("</TD><TD></TD>\n");
};

void html_print_moon()
{
    signed char i;

    printf("<TD>");

    if (wg_flags & wg_fl_earth)
    {
        i = earth_moon_phase(current_day, current_month, current_year);
        printf("Luna: ");
    }
    else
    {
        i = harn_moon_phase(current_day, current_month, current_year);
        printf("Yael: ");
    }

    printf("</TD><TD>");

    if (i == 0)
        wg_moon_age = 0;
    else
        wg_moon_age++;

    switch(i)
    {
        case -100:  printf("Full Moon");        break;
        case 50:    printf("First Quarter");    break;
        case -50:   printf("Last Quarter");     break;
        case 0:     printf("New Moon");         break;
        default:
            if (wg_flags & wg_fl_moon_percent)
            {
                if (i > 50)
                    printf("Wax. Gibbous&nbsp; (%2i%%)", i);
                else if (i > 0)
                    printf("Wax. Crescent (%2i%%)", i);
                else if (i < -50)
                    printf("Wan. Gibbous&nbsp; (%2i%%)", -i);
                else
                    printf("Wan. Crescent (%2i%%)", -i);
            }
            else if (wg_flags & wg_fl_moon_age)
            {
                if (i > 50)
                    printf("Wax. Gibbous&nbsp;");
                else if (i > 0)
                    printf("Wax. Crescent");
                else if (i > -50)
                    printf("Wan. Crescent");
                else
                    printf("Wan. Gibbous&nbsp;");

                printf(" (%2id)", wg_moon_age);
            }
            else
            {
                if (i > 0)
                    printf("Waxing");
                else
                {
                    i = -i;
                    printf("Waning");
                }

                if (i > 50)
                    printf(" Gibbous");
                else
                    printf(" Crescent");
            }
    }

    printf("</TD><TD></TD>\n");
};

void html_print_newline()
{
    printf("</TR><TR>\n");
};

void html_print_weekday()
{
    if (wg_flags & wg_fl_earth)
        printf("<TD>&nbsp;</TD><TD align=\"right\">%s</TD><TD></TD>\n", earth_weekday(current_day, current_month, current_year));
    else
        printf("<TD>&nbsp;</TD><TD>&nbsp;</TD><TD></TD>\n");
};

void html_print_newday()
{
    printf("<TABLE border=3 cellspacing=0 cellpadding=2 width=\"100%%\" valign=\"top\">\n");
};

void html_print_newmonth(char *c_month)
{
    printf("\n<BR><H3>");

    if (!strcmp(c_month, "Kelen"))
        printf("Kelen/Saraju");
    else
        printf("%s", c_month);

    printf("<HR size=1 width=\"100%%\"></H3>\n");
};

/*****************************************************************************
  * Some silly functions to call functions from another module               *
  ****************************************************************************/

unsigned char h_get_season(signed char month, signed long year)
{
    return (harn_get_season(month, year));
};

unsigned char e_get_season(signed char month, signed long year)
{
    return (earth_get_season(month, year));
};

unsigned char h_num_of_days(signed char month, signed long year)
{
    return (harn_num_of_days(month, year));
};

unsigned char e_num_of_days(signed char month, signed long year)
{
    return (earth_num_of_days(month, year));
};

/*****************************************************************************
  * sun_times: Get the rise and set times of the sun                         *
  ****************************************************************************/

void sun_times(signed char *r_h, signed char *r_m, signed char *s_h, signed char *s_m)
{
    signed char j, rs;
    float rise, set;

    if (wg_flags & wg_fl_earth)
        rs = sun_rise(current_day, current_month, current_year, wg_latitude, earth_orbit_data, &rise, &set);
    else
        rs = sun_rise(current_day, current_month, current_year, wg_latitude, kethira_orbit_data, &rise, &set);

    switch( rs )
    {
        case +1:
                *r_h = 0;
                *r_m = 0;
                *s_h = 24;
                *s_m = 0;
            break;
        default:
                *r_h = (int)floor(rise);
                *r_m = (int)( (float)( (rise - floor(rise)) * 100.0) * 0.6);

                *s_h = (int)floor(set);
                *s_m = (int)( (float)( (set - floor(set)) * 100.0) * 0.6);
    }
};

/*****************************************************************************
  * is_night_watch: See if the watch occurs at night                         *
  ****************************************************************************/

int is_night_watch(signed char num, signed char r_h, signed char r_m, signed char s_h, signed char s_m)
{
    while (num > 6)
        num -= 6;
    while (num < 1)
        num += 6;

        /*  Get the middle hour of the watch (2, 6, 10, 14, 18, 22)  */

    num = ((num - 1) * 4) + 2;

        /* If the sun rises after that, it is a night watch */

    if ((float)(r_h + (r_m / 60.0)) > (float)num)
        return 1;

        /* ...or, perhaps it has already set by then?       */

    if ((float)(s_h + (s_m / 60.0)) < (float)num)
        return -1;

        /* If not, then it must be daytime                  */

    return 0;
};

/*****************************************************************************
  * DJGPP: Two empty functions to cut down on executable size                *
  ****************************************************************************/

#ifdef DJGPP
#include <crt0.h>

char **__crt0_glob_function(char *_argument)        { return 0; };
void __crt0_load_environment_file(char *_app_name)  { return;   };

#endif
